/* The following copy right applies to all the code in this file except: */
/* GetCannonicalInfo()  		 	 			 */

/*
 * Copyright (c) 2002
 * Ningning Hu and the Carnegie Mellon University.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author(s) may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* The following copyright applies to the following procedures: 	*/
/* GetCannonicalInfo()  						*/

/*
 * Copyright (c) 1999
 * Stefan Savage and the University of Washington.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author(s) may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h> 
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>

#include "setsignal.h"
#include "common.h"

#define PhaseSleep	(0)
#define ABS(a)     (((a) < 0) ? -(a) : (a))

extern char *optarg;	
extern int optind, opterr, optopt;

char version[] = "1.0";

int delay_num = 0;
int packet_size = PacketSize;
int probe_num = ProbeNum;
FILE * trace_fp = NULL;
int verbose = 0;
int debug = 0;
char work_mode[16];

double b_bw = 0;
double competing_bw, PTR_bw, a_bw;
int probing_phase_count = 0;
double probing_start_time = 0;
double probing_end_time = 0;

uint16 dst_port = START_PORT;
uint16 src_port = 0;
uint16 probing_port;
char dst_hostname[MAXHOSTNAMELEN];
uint32 dst_ip;
char dst_ip_str[16];
char src_hostname[MAXHOSTNAMELEN];
uint32 src_ip;
char src_ip_str[16];
char dst[MAXHOSTNAMELEN];
char src[MAXHOSTNAMELEN];
int control_sock;
int probing_sock;
struct sockaddr_in probing_server;

RETSIGTYPE (*oldhandler)(int);

double src_gap_sum = 0; // entry value
double avr_src_gap; // entry value
struct pkt_rcd_t dst_gap[MaxProbeNum];
double dst_gap_sum;
double avr_dst_gap;
int dst_gap_count, src_gap_count;
int total_count;
double send_times[MaxProbeNum];
/* "1" means good gap, "0" denotes a bad gap starting with the packet */
int packet_sanity[MaxProbeNum];

/* keep the data from dst machine, 128 is for other control string */
char msg_buf[sizeof(struct pkt_rcd_t) * MaxProbeNum + 128];
int  msg_len;

/* trace item: record the item used to dump out into trace file */
struct trace_item {
	int probe_num, packet_size, delay_num;

     	double send_times[MaxProbeNum];
	int packet_sanity[MaxProbeNum];
	struct pkt_rcd_t rcv_record[MaxProbeNum];
	int record_count;

	double avr_src_gap;
	double avr_dst_gap;

	double b_bw, c_bw, a_bw, ptr;

	struct trace_item * next;
};
struct trace_item * trace_list = NULL;
struct trace_item * trace_tail = NULL;
struct trace_item * cur_trace = NULL;


/**********************************************************************/

/* usage message */
void Usage()
{
printf("igi-%s client usage:\n\n", version);
printf("\tigi_client [-hvd] [-n probe_num] [-s packet_size] [-p dst_port]\n");
printf("\t           [-m working_mode] [-f trace_file] dst_address\n\n");
printf("\t-h      print this message.\n");
printf("\t-v      verbose mode.\n");
printf("\t-d      debug mode.\n");
printf("\t-n      set the number of probing packets in each phase.\n");
printf("\t-s      set the length(B) of the probing packets in byte.\n");
printf("\t-p      indicate the dst machine's listening port.\n");
printf("\t        This is optional, it can itself search for the\n");
printf("\t        port that the igi_server is using.\n");
printf("\t-m      working mode:\n");
printf("\t        \"simple\": get the result just using 2 phases of probing\n");
printf("\t        \"binary\": using binary search to get the final result\n");
printf("\t        \"sequential\": sequentially search for the final result\n");
printf("\t-f      indicate the trace file name, which records the packet\n");
printf("\t        information in way that matlab script can use.\n\n");
printf("dst_address     can be either IP address or hostname of dst machine.\n\n");

exit(1);
}


/* combine the sec & usec of record into a real number */
double get_rcd_time(struct pkt_rcd_t record)
{
	return (record.sec + (double)record.u_sec / 1000000);
}


/* dump out all the trace packet time stamps */
void dump_trace()
{
	struct trace_item * p;
	int i, index;

	if (trace_fp == NULL) {
	    printf("-w not specified, no trace to dump\n");
	    return;
	}

	/* first dump out the summary data */
	p = trace_list;
	index = 0;
	fprintf(trace_fp, "\n%%probe_num packet_size delay_num avr_src_gap arv_dst_gap b_bw c_bw a_bw ptr\n");
	fprintf(trace_fp, "summary_data = [\n");
	while (p != NULL) {
	    index ++;
	    fprintf(trace_fp,"%2d %4d %5d %f %f %12.3f %12.3f %12.3f %12.3f\n", 
			    p->probe_num,
			    p->packet_size,
			    p->delay_num,
			    p->avr_src_gap,
			    p->avr_dst_gap,
			    p->b_bw,
			    p->c_bw,
			    p->a_bw,
			    p->ptr);
	    p = p->next;
	}
	fprintf(trace_fp, "];\n");
	fprintf(trace_fp, "phase_num = %d; \n\n", index);

	/* then the detail packet trace */
	p = trace_list;
	index = 0;
	while (p != NULL) {
	    index ++;

	    fprintf(trace_fp, "send_time_%d = [\n", index);
	    for (i=1; i<p->probe_num; i++) {
		/*fprintf(trace_fp, "%f %f %f ", 
			p->send_times[i-1], p->send_times[i], 
			p->send_times[i] - p->send_times[i-1]);*/
		fprintf(trace_fp, "%f ", 
			p->send_times[i] - p->send_times[i-1]);
		if (p->packet_sanity[i-1])
		    fprintf(trace_fp, "1 \n");
		else
		    fprintf(trace_fp, "0 %% * \n");

	    } 
	    fprintf(trace_fp, "];\n");
	    fprintf(trace_fp, "send_array_size(%d) = %d; \n\n", 
			    index, p->probe_num - 1);


	    fprintf(trace_fp, "recv_time_%d = [\n", index);

	    for (i=0; i<p->record_count-1; i++) {
	    	/*fprintf(trace_fp, "%f %f %f %d ", 
			get_rcd_time(p->rcv_record[i]),
			get_rcd_time(p->rcv_record[i+1]),
			get_rcd_time(p->rcv_record[i+1]) - 
			get_rcd_time(p->rcv_record[i]),
			p->rcv_record[i+1].seq);*/
	    	fprintf(trace_fp, "%f %d ", 
			get_rcd_time(p->rcv_record[i+1]) - 
			get_rcd_time(p->rcv_record[i]),
			p->rcv_record[i+1].seq);

		if (p->packet_sanity[p->rcv_record[i].seq])
		    fprintf(trace_fp, "1 \n");
		else
		    fprintf(trace_fp, "0 %% * \n");
	    }
	    fprintf(trace_fp, "];\n");
	    fprintf(trace_fp, "recv_array_size(%d) = %d; \n\n", 
			    index, p->record_count - 1);

	    p = p->next;
	}
}


/* make a clean exit on interrupts */
RETSIGTYPE cleanup(int signo)
{
	if (trace_fp != NULL) {
	    dump_trace();
	    fclose(trace_fp);
	}

	close(probing_sock);
	close(control_sock);

	/* free trace list */
  	if (trace_list != NULL) {
	    struct trace_item * pre, * p;
	    p = trace_list;

	    while (p != NULL) {
		pre = p;
		p = p->next;
		free(pre);
	    }
  	}

  	exit(0);
}


/* copied from sting.c */
void Quit()
{
  fflush(stdout);
  fflush(stderr);
  cleanup(1);
  exit(-1);
}


/* get the current time */
double get_time()
{
        struct timeval tv;
        struct timezone tz;
        double cur_time;

        if (gettimeofday(&tv, &tz) < 0) {
            perror("get_time() fails, exit\n");
            Quit();
        }

        cur_time = (double)tv.tv_sec + ((double)tv.tv_usec/(double)1000000.0);
        return cur_time;
}


/* Init session.
 * -- modified from sting.c's EstablishSession() */
int init_session(uint32 sourceAddress, uint16 sourcePort,
		 uint32 targetAddress, uint16 targetPort)
{
	struct sockaddr_in client;

	probing_sock = socket(AF_INET, SOCK_DGRAM, 0);

	probing_server.sin_family = AF_INET;
	probing_server.sin_addr.s_addr = targetAddress;
	probing_server.sin_port = htons(probing_port);

	client.sin_addr.s_addr = htonl(INADDR_ANY);
	client.sin_family = AF_INET;
	client.sin_port = htons(0);

        if (bind(probing_sock, (const struct sockaddr *) &client,
                 sizeof(client)) < 0) {
           fprintf(stderr, "server: Unable to bind local address.\n");
           Quit();
        }

#ifndef LINUX
        if (fcntl(probing_sock, F_SETFL, O_NDELAY) < 0) {
#else
        if (fcntl(probing_sock, F_SETFL, FNDELAY) < 0) {
#endif
           printf("fcntl fail \n");
           Quit();
        }

	return 1;
}


/* get ip adress, hostname, and the corresponding sockaddr struct */
/* -- copied from sting.c                                         */
int GetCannonicalInfo(char *string, 
		      char name[MAXHOSTNAMELEN], 
		      uint32 *address, 
		      char ip_str[16])
{
  struct hostent *hp;
  struct in_addr in_a;

  /* Is string in dotted decimal format? */
#ifndef LINUX
  if ((*address = inet_addr(string)) == INADDR_BROADCAST) {
#else
  if ((*address = inet_addr(string)) == INADDR_NONE) {
#endif
    /* No, then lookup IP address */
    if ((hp = gethostbyname(string)) == NULL) {
      /* Can't find IP address */
      printf("ERROR: couldn't obtain address for %s\n", string);
      return -1;
    } else {
      strncpy(name, hp->h_name, MAXHOSTNAMELEN-1);
      memcpy((void *)address, (void *)hp->h_addr, hp->h_length);
      in_a.s_addr = *address;
      strcpy(ip_str, inet_ntoa(in_a)); 
    }
  } else {
    strcpy(ip_str, string);
    if ((hp = gethostbyaddr((char *)address, sizeof(*address),
			    AF_INET)) == NULL) {
      /* Can't get cannonical hostname, so just use input string */
      if (verbose) {
	printf("WARNING: couldn't obtain cannonical name for %s\n", string);
      }
      strncpy(name, string, MAXHOSTNAMELEN - 1);
    } else {
      strncpy(name, hp->h_name, MAXHOSTNAMELEN - 1);
    }
  }
  return 0;
}


/* send out probe_num packets, each packet is packet_size bytes,   */
/* and the inital gap is set using delay_num                       */
void send_packets( 
	     int probe_num,
	     int packet_size, 
	     int delay_num,
	     double *sent_times)
{
	int i, k;
	double tmp = 133333.0003333;
	char send_buf[4096];

	/* send out probing packets */
	for (i=0; i<probe_num; i++) {
	    sent_times[i] = get_time();
	    // TODO: this can only up to 127
	    send_buf[0] = i;
	    sendto(probing_sock, send_buf, packet_size, 0,
	                        (struct sockaddr *)&(probing_server), 
				sizeof(probing_server));

	    // TODO: any better ways to create small time delay?
	    for (k=0; k<delay_num; k++) {
		tmp = tmp * 7;
		tmp = tmp / 13;
	    }
	}
}


/* the begining of the msg_buf should have the form $**$, return the  */
/* string "**" 							      */
// TODO: simplify using index()
void get_item(int control_sock, char *str)
{
	int i, j;

	/* check if two $ already exit */
	if (msg_len > 0) {
	    if (msg_buf[0] != '$') {
		printf("get unknown msg, it will mess up everything, exit\n");
		Quit();
	    }
	    i=1;
	    while ((i<msg_len) && (msg_buf[i] != '$')) {
		i ++;
	    }
	    if (i<msg_len) {
		strncpy(str, msg_buf+1, i-1);
		str[i-1] = 0;
		goto out;
	    }
	}

	/* read until we get two $ */
        while (1) {
	    if (debug) {
	        printf("msg_len = %d  ", msg_len);
	    }
            msg_len += read(control_sock, msg_buf+msg_len, 
			    sizeof(msg_buf) - msg_len);
	    msg_buf[msg_len] = 0;
	    // printf("msg_buf = %s\n", msg_buf);

	    i = 1; 
	    while ((i<msg_len) && (msg_buf[i] != '$')) {
		i ++;
	    }
	    if (i<msg_len) {
		strncpy(str, msg_buf+1, i-1);
		str[i-1] = 0;
		break;
	    }
        }

out:
	/* get out the string and move the buf ahead, with i points
	 * to the 2nd $ */
	i++;
	for (j=i; j<msg_len; j++) {
	     msg_buf[j-i] = msg_buf[j];
	}
	msg_len -= i;
}


/* get dst gap_sum and gap_count from the records */
double get_dst_sum(struct pkt_rcd_t *rcv_record, 
		   int count, 
		   int *gap_count)
{
        double gap_sum;
	double time1, time2;
        int i;

        gap_sum = 0;
        *gap_count = 0;
        for (i=0; i<count-1; i++) {
            /* only consider those gaps composed by two in-order packets */
	    if ((rcv_record[i+1].seq - rcv_record[i].seq == 1)
		&& (packet_sanity[rcv_record[i].seq] == 1))  {
		time1 = get_rcd_time(rcv_record[i+1]);
		time2 = get_rcd_time(rcv_record[i]);
                gap_sum += (time1 - time2);
                (*gap_count) ++;

	    	// printf("%f %f %f \n", time1, time2, time1 - time2);
                // printf("gap_count = %d\n", *gap_count);
            }
        }

        return gap_sum;
}

/* get dst gap trace data through control channel 	*/
/* return 0 means fails in reading dst's data           */
int get_dst_gaps(struct pkt_rcd_t * rcv_record)
{
	int i;

	char num_str[32];
	int  data_size; 

	int  total_count;
	int  ptr;

	get_item(control_sock, num_str);
	data_size = atoi(num_str); // this depends on stdlib.h
	total_count = data_size / sizeof(struct pkt_rcd_t);
	if (verbose) {
	    printf("from dst: data_size = %d total_count = %d \n", 
			    data_size, total_count);
	}

	/* get the data package */
	while (msg_len < data_size) {
            msg_len += read(control_sock, msg_buf+msg_len, 
			    sizeof(msg_buf) - msg_len);
	}

	/* split out the record */
	ptr = 0;
	for (i=0; i<total_count; i++) {
	    memcpy((void *)&rcv_record[i], 
		   (const void *)(msg_buf + ptr), 
		   sizeof(struct pkt_rcd_t));

	    rcv_record[i].sec = ntohl(rcv_record[i].sec);
	    rcv_record[i].u_sec = ntohl(rcv_record[i].u_sec);
	    rcv_record[i].seq = ntohl(rcv_record[i].seq);

	    if (debug) {
	    	printf("%d %d %d \n", 
		    rcv_record[i].sec, 
		    rcv_record[i].u_sec, 
		    rcv_record[i].seq);
	    }
	    ptr += sizeof(struct pkt_rcd_t);
	}
	msg_len -= data_size;

	if (trace_fp != NULL) {
	    memcpy(cur_trace->rcv_record, rcv_record, data_size);
	    cur_trace->record_count = total_count;
	}

	return total_count;
}


/* get the turning point posistion, return its array index*/
int get_turning_point(double * src_gap,
		      double * dst_gap,
		      int count)
{
	int i = 0;

	while (i<count) {
	    /* throw away the bad data */
	    if ((src_gap[i] < 0.000005) || (dst_gap[i] == 0.000005)) {
		i++;
	    	continue;
	    }

	    if (ABS(dst_gap[i] - src_gap[i]) < dst_gap[i] * 0.05) 
		break;

	    i++;
	}

	return i;
}


/* here we use some a similar technique with nettimer, but the  */
/* implementation is much simpler 			        */
double get_bottleneck_bw(struct pkt_rcd_t *rcv_record, int count)
{
	double gaps[MaxProbeNum];
	int gap_count;

	struct bin_item {
	    int value;
	    int count;
	} bin[MaxProbeNum];
	int bin_count;

	double lower_time, upper_time;
	int max_index, max_count, max_value, cur_value;
	int i, j;

	double gap_sum;
	int sum_count;

	/* get valid dst gaps */
	gap_count = 0;
	for (i=0; i<count-1; i++) {
	    if ((rcv_record[i+1].seq - rcv_record[i].seq == 1) 
		&& (packet_sanity[rcv_record[i].seq] == 1)) {
	        gaps[gap_count] = get_rcd_time(rcv_record[i+1]) 
			- get_rcd_time(rcv_record[i]);
		gap_count ++;
	    }
	}

	/* use 25us bins to go through all the gaps */
	bin_count = 0;
	for (i=0; i<gap_count; i++) {
	    cur_value = (int)(gaps[i]/BinWidth);
	    j = 0;
	    while ((j<bin_count) && (cur_value != bin[j].value))  {
		j++;
	    }
	    if (j == bin_count) {
		bin[bin_count].value = cur_value;
		bin[bin_count].count = 1;
		bin_count ++;
	    }
	    else {
		bin[j].count ++;
	    }
	}

	/* find out the biggest bin */
	max_index = 0; max_count = bin[0].count;
	for (i=1; i<bin_count; i++) {
	    if (bin[i].count > max_count) {
		max_count = bin[i].count;
		max_index = i;
	    }
	}
	max_value = bin[max_index].value;
	lower_time = max_value * BinWidth;
	upper_time = (max_value + 1) * BinWidth;

	/* see whether the adjacent two bins also have some items */
	for (i=0; i<bin_count; i++) {
	    if (bin[i].value == max_value + 1) 
		upper_time += BinWidth;
	    if (bin[i].value == max_value - 1) 
		lower_time -= BinWidth;
	}

	if (debug) {
	    printf("get_bottleneck_bw: %f %f \n", lower_time, upper_time);
	}

	/* average them */
	gap_sum = 0;
	sum_count = 0;
	for (i=0; i<gap_count; i++) {
	    if ((gaps[i] >= lower_time) && (gaps[i] <= upper_time)) {
	      	gap_sum += gaps[i];
	    	sum_count ++;
	    }
	}

	return (packet_size * 8 / (gap_sum / sum_count));
}


/* calculate the competing traffic rate using IGO method */
double get_competing_bw(struct pkt_rcd_t *rcv_record, 
			double avr_src_gap,
			int count, 
			double b_bw)
{
	double b_gap, cur_gap, m_gap;
	double gap_sum, inc_gap_sum;
        int i;

	b_gap = packet_size * 8 / b_bw;
	m_gap = MAX(b_gap, avr_src_gap);

	gap_sum = inc_gap_sum = 0;
        for (i=0; i<count-1; i++) {
            /* only consider those gaps composed by two in-order packets */
            if ((rcv_record[i+1].seq - rcv_record[i].seq == 1) 
		&& (packet_sanity[rcv_record[i].seq])) {
		cur_gap = get_rcd_time(rcv_record[i+1]) 
			- get_rcd_time(rcv_record[i]);
		gap_sum += cur_gap;

		if (cur_gap > m_gap + 0.000005)
                    inc_gap_sum += (cur_gap - b_gap);

		/*if (debug) {
		    printf("%d %d %d %d %d %d\n", 
			rcv_record[i].sec,
			rcv_record[i].u_sec,
			rcv_record[i].seq,
			rcv_record[i+1].sec,
			rcv_record[i+1].u_sec,
			rcv_record[i+1].seq);
		}*/
            }
        }

	if (debug) {
	    printf("b_gap = %f %f %f \n", b_gap, inc_gap_sum, gap_sum);
	}

	if (gap_sum == 0) {
	    printf("\nNo valid trace in the last phase probing\n");
	    printf("You may want to try again. Exit\n");
	    Quit();
	}

        return (inc_gap_sum * b_bw / gap_sum);
}


/* check if two gap values can be considered "equal" */
/* invalid values are considered as high	     */
int gap_comp(double dst_gap, double src_gap)
{
	double delta;

	delta = dst_gap * GapDiffIgnoreRatio;
	if (delta < GapDiffIgnoreValue) 
	    delta = GapDiffIgnoreValue; 

	/* sanity check */
	if ((dst_gap == 0) || (src_gap == 0))
	    return 1;

	/* "dst_gap is not supposed to be smaller src_gap"
	 * this needs more research, but up to now this is a 
	 * valid assumption for us */
	if (dst_gap < src_gap - delta) {
	    if (verbose) {
	       	printf("smaller dst_gap, considered to be equal \n");
	    }
	    return -1;
	}

	if (dst_gap > src_gap + delta)
	    return 1;
	else 
	    return 0;
}


/* flag those packets which have "anormal" src gaps, that is too small 	*/
/* or too big, and will not use them in the future calcualtion. 	*/
/* The algorithm is the same as that used in get_bottleneck_bw() 	*/
void sanity_check(double * times, int * gap_count)
{
	double gaps[MaxProbeNum];

	struct bin_item {
	    int value;
	    int count;
	} bin[MaxProbeNum];
	int bin_count;

	int cur_value;
	int max_index, max_count, max_value;
	double lower_time, upper_time;
	int i, j;

	/* initialize */
	for (i=0; i<MaxProbeNum; i++) {
	    packet_sanity[i] = 0; 
	}
	for (i=0; i<probe_num-1; i++) {
	    gaps[i] = times[i+1] - times[i];
	}

	/* use 25us bins to go through all the time stamps*/
	bin_count = 0;
	for (i=0; i<probe_num-1; i++) {
	    cur_value = (int)(gaps[i]/BinWidth);
	    j = 0;
	    while ((j<bin_count) && (cur_value != bin[j].value))  {
		j++;
	    }
	    if (j == bin_count) {
		bin[bin_count].value = cur_value;
		bin[bin_count].count = 1;
		bin_count ++;
	    }
	    else {
		bin[j].count ++;
	    }
	}

	/* find out the biggest bin */
	max_index = 0; max_count = bin[0].count;
	for (i=1; i<bin_count; i++) {
	    if (bin[i].count > max_count) {
		max_count = bin[i].count;
		max_index = i;
	    }
	}
	max_value = bin[max_index].value;
	lower_time = max_value * BinWidth;
	upper_time = (max_value + 1) * BinWidth;

	/* see whether the adjacent two bins also have some items */
	for (i=0; i<bin_count; i++) {
	    if (bin[i].value == max_value + 1) 
		upper_time += BinWidth;
	    if (bin[i].value == max_value - 1) 
		lower_time -= BinWidth;
	}

	if (debug) {
	    printf("%f %f \n", lower_time, upper_time);
	}

	/* check the packet sanity */
	*gap_count = 0;
	for (i=0; i<probe_num-1; i++) {
	    // printf("%f \n", gaps[i]);
	    if ((gaps[i] < upper_time) && (gaps[i] > lower_time)) {
		packet_sanity[i] = 1;
		(*gap_count) ++;
	    }
	    else {
		packet_sanity[i] = 0;
	    }
	}
}


/* return summary of the "valid" src gaps */
double get_src_sum(double * times)
{
	int i;
	double sum;

	sum = 0;
	for (i=0; i<probe_num-1; i++) {
	    // printf("%f %f %f \n", times[i], times[i+1], times[i+1] - times[i]);
	    if (packet_sanity[i]) 
		sum += (times[i+1] - times[i]);
	}

	return sum;
}


double get_bandwidth()
{
	if (dst_gap_count == 0) 
	    competing_bw = PTR_bw = a_bw = 0;
	else {
	    /* bottleneck bandwidth */
	    // b_bw = packet_size * 8 / avr_dst_gap; 
	    if (b_bw < 1) 
	        b_bw = get_bottleneck_bw((struct pkt_rcd_t *)&dst_gap, 
				         total_count);

	    /* PTR throughput */
	    PTR_bw = packet_size * dst_gap_count * 8 / dst_gap_sum;

	    /* competing traffic rate */
	    competing_bw = get_competing_bw(
			(struct pkt_rcd_t *)&dst_gap, 
			avr_src_gap,
			total_count, 
			b_bw);

	    /* available bandwidth */
	    a_bw = b_bw > competing_bw ? b_bw - competing_bw : 0;
	}

	return a_bw;
}


/* one complete probing procedure, which includes:			*/
/* 1. send out packets with (probe_num, packet_size, delay_num)  	*/
/* 2. retrieve the dst gaps values from dst machine			*/ 
/* 3. filter out those bad trace data and compute the src gap & dst 	*/ 
/*    gap summary and average						*/ 
void one_phase_probing()
{
	if (verbose) {
	    printf("\nprobe_num = %d packet_size = %d delay_num = %d \n", 
		    probe_num, packet_size, delay_num);
	}

	/* probing */
  	send_packets(probe_num, packet_size, delay_num, send_times);
	if (trace_fp != NULL) {
	    /* create a new trace item */
	    cur_trace = (struct trace_item *)malloc(sizeof(struct trace_item));
	    cur_trace->next = NULL;
	    if (trace_list == NULL) {
		trace_list = cur_trace;
		trace_tail = cur_trace;
	    }
	    else {
	        trace_tail->next = cur_trace;
	        trace_tail = cur_trace;
	    }
	}
	/* get dst gap logs */
	total_count = get_dst_gaps((struct pkt_rcd_t *)&dst_gap);

	/* get src & dst gap statistics */
	sanity_check(send_times, &src_gap_count); 
	if (debug) {
	    printf("src_gap_count = %d \n", src_gap_count);
	}
	src_gap_sum = get_src_sum(send_times);
	// src_gap_sum = send_times[probe_num-1] - send_times[0];
	if (src_gap_count != 0) 
	    avr_src_gap = src_gap_sum / src_gap_count;
	else 
	    avr_src_gap = 0;

	dst_gap_sum = get_dst_sum(
			    (struct pkt_rcd_t *)&dst_gap, 
			    total_count, 
			    &dst_gap_count);
	if (debug) {
	    printf("%d dst_gap_count = %d \n", total_count, dst_gap_count);
	}
	if (dst_gap_count != 0)
	    avr_dst_gap = dst_gap_sum / dst_gap_count;
	else
	    avr_dst_gap = 0;

	if (verbose) {
	    printf("avr_src_gap = %f \n", avr_src_gap);
	    printf("avr_dst_gap = %f \n", avr_dst_gap);
	}

	if (trace_fp != NULL) {
	    double tmp1, tmp2, tmp3;

	    cur_trace->probe_num = probe_num;
	    cur_trace->packet_size = packet_size;
	    cur_trace->delay_num = delay_num;

	    memcpy(cur_trace->send_times, send_times, 
			    sizeof(double) * probe_num);

	    memcpy(cur_trace->packet_sanity, packet_sanity, 
			    sizeof(int) * probe_num);

	    cur_trace->avr_src_gap = avr_src_gap;
	    cur_trace->avr_dst_gap = avr_dst_gap;

	    /* we can't use get_bandwidth() here since we want to record
	     * the current data, which may be wrong value*/
	    if (dst_gap_count == 0)
		cur_trace->b_bw = 0;
	    else {
	        cur_trace->b_bw = get_bottleneck_bw((struct pkt_rcd_t *)&dst_gap, 
				         total_count);
	    }

	    tmp1 = PTR_bw;
	    tmp2 = competing_bw;
	    tmp3 = a_bw;
	    get_bandwidth();
	    cur_trace->ptr = PTR_bw;
	    cur_trace->c_bw = competing_bw;
	    cur_trace->a_bw = a_bw;
	    PTR_bw = tmp1;
	    competing_bw = tmp2;
	    a_bw = tmp3;
	}

	probing_phase_count ++;
}


/* find the delay number which can set the src_gap exactly as "gap" */
int get_delay_num(double gap) 
{
#define Scale 		(10)
	int lower, upper, mid;
	double s_time, e_time, tmp;
	int k;

	lower = 0;
	upper = 16;
	tmp = 133333;

	/* search for upper bound */
	s_time = e_time = 0;
	while (e_time - s_time < gap * Scale) {
	    s_time = get_time();
	    for (k=0; k<upper * Scale; k++) {
		tmp = tmp * 7;
		tmp = tmp / 13;
	    }
	    e_time = get_time();

	    upper *= 2;
	}

	/* binary search for delay_num */
	mid = (int)(upper + lower) / 2;
	while (upper - lower > 20) {
	    s_time = get_time();
	    for (k=0; k<mid * Scale; k++) {
		tmp = tmp * 7;
		tmp = tmp / 13;
	    }
	    e_time = get_time();

	    if (e_time - s_time > gap * Scale) 
		upper = mid;
	    else 
		lower = mid;

	    mid = (int)(upper + lower) / 2;
	}

	return mid;
}


void dump_bandwidth()
{
	printf("\n");
	printf("Bottleneck Bandwidth: %f \n", b_bw);
	printf("Competing  Bandwidth: %f \n", competing_bw);
	printf("Packet Transmit Rate: %f \n", PTR_bw);
	printf("Available  Bandwidth: %f \n", a_bw);

	if (trace_fp != NULL) {
	    fprintf(trace_fp, "%%Bottleneck Bandwidth: %f \n", b_bw);
	    fprintf(trace_fp, "%%Competing  Bandwidth: %f \n", competing_bw);
	    fprintf(trace_fp, "%%Packet Transmit Rate: %f \n", PTR_bw);
	    fprintf(trace_fp, "%%Available  Bandwidth: %f \n", a_bw);
	}
}

void dump_stat()
{
	printf("\n%s probing time : %f\n", 
			work_mode, probing_end_time - probing_start_time);
	printf("%s probing count: %d\n", work_mode, probing_phase_count);
	printf("ending gaps: [ %f %f ]\n", avr_src_gap, avr_dst_gap);
}


/* This method simply assump that the turning point is at the bottleneck gap,
 * and just do two probings:
 * 1st: get the b_gap
 * 2nd: probe on b_gap (turing point)
 * It is not very accurate, but much faster. */
double simple_probing()
{
	// double turning_gap;
	double my_bw;

	one_phase_probing();
	// turning_gap = avr_dst_gap;
	b_bw = get_bottleneck_bw((struct pkt_rcd_t *)&dst_gap, total_count);

	delay_num = get_delay_num(avr_dst_gap);
	one_phase_probing();

	my_bw = get_bandwidth();
	// b_bw = packet_size * 8 / turning_gap; 
	dump_bandwidth();

	return my_bw;
}


/* Using binary search to locate the turning point */
double binary_probing()
{
	int lower_delay, upper_delay;
	double turning_gap;
	double my_b_bw, my_a_bw;
	int try_count = 0;
	
	if (verbose) {
	    printf("start binary probing\n");
	}

	/* Lower bound is set with 0, but we need to confirm it */
	lower_delay = 0;
	delay_num = 0;
	one_phase_probing();
	try_count = 1;
	while ((try_count <= 3) && (avr_dst_gap < avr_src_gap)) {
	    try_count ++;
	    one_phase_probing();
	}

	if (avr_dst_gap < avr_src_gap) {
	    printf("Bottleneck link capacity too high. Can not use this tool.\n");
	    Quit();
	}
	/* the bottleneck bandwidth is no longer based on turning point 
	 * here any more */
	my_b_bw = get_bottleneck_bw((struct pkt_rcd_t *)&dst_gap, total_count);
	b_bw = my_b_bw;

	//TODO: is these sleeps really necessary?
	sleep(PhaseSleep);

	/* locate the upper bound */
	delay_num = get_delay_num(avr_dst_gap);
	one_phase_probing();
	while (gap_comp(avr_dst_gap, avr_src_gap) > 0) {
	    if (delay_num >= BigDelay) {
		lower_delay = delay_num;
	    }

	    delay_num *= 2;	
	    one_phase_probing();
	    sleep(PhaseSleep);

	}
	upper_delay = delay_num;

	/* narrow down */
	/* TODO: this is the research about turning point */
	while (upper_delay >= lower_delay + SmallDelay) {
	    if (debug) {
	        printf("\nbound: %d %d \n", lower_delay, upper_delay);
	    }
	    delay_num = (int)((upper_delay + lower_delay) / 2);
	    one_phase_probing();

	    if (gap_comp(avr_dst_gap, avr_src_gap) > 0) 
		lower_delay = delay_num;
	    else
		upper_delay = delay_num;

	    sleep(PhaseSleep);
	    
	}
	turning_gap = avr_dst_gap;
	if (verbose) {
	    printf("Turning gap = %f\n", turning_gap);
	}
	if (trace_fp != NULL) {
	    fprintf(trace_fp, "%%Turning gap = %f \n", turning_gap);
	}
	
	my_a_bw = get_bandwidth();
	dump_bandwidth();

	return my_a_bw;
}


/* This method tries to achieve the best speed to get the final results. 
 * The following techniques are taken to do this:
 * 1. no sleep;
 * 2. jump to b_gap directly, and use b_gap to adjust the source gap */
double fast_probing()
{
	double interval;
	int end_num;
	double my_a_bw = 0;

	probing_start_time = get_time();

	delay_num = 0;
	one_phase_probing();

	delay_num = get_delay_num(avr_dst_gap);

	interval = (double) delay_num / 8;
	end_num = 2 * delay_num;

	delay_num /= 2;

	delay_num = (int)(delay_num + interval);
	one_phase_probing();
	b_bw = get_bottleneck_bw((struct pkt_rcd_t *)&dst_gap, total_count);

	while (delay_num <= end_num) {
	    delay_num = (int)(delay_num + interval);
	    one_phase_probing();

	    if ((my_a_bw == 0) 
		&& (gap_comp(avr_dst_gap, avr_src_gap) == 0)) {
		my_a_bw = get_bandwidth();
	 	if (trace_fp != NULL) {
	    	    fprintf(trace_fp, "%%Seqential mode: get the value when delay_num = %d \n\n", delay_num); 
		}
		break;
	    }

	}

	if (my_a_bw == 0) {
	    do {
	    	delay_num = (int)(delay_num + interval);
	    	one_phase_probing();
	    } while (gap_comp(avr_dst_gap, avr_src_gap) != 0);
	    my_a_bw = get_bandwidth();
	    if (trace_fp != NULL) {
	        fprintf(trace_fp, 
		"%%Seqential mode: get the value when delay_num = %d \n\n", 
		delay_num); 
	    }
	}

	dump_bandwidth();

	probing_end_time = get_time();
	dump_stat();

	return my_a_bw;
}

/* This just starts from a very low src gap, sequentially increases it, */
/* until the src gap is twice of the possible turning gap value		*/
double sequential_probing()
{
	double interval;
	int end_num;
	double my_a_bw = 0;

	probing_start_time = get_time();

	delay_num = 0;
	one_phase_probing();
	sleep(PhaseSleep);

	delay_num = get_delay_num(avr_dst_gap);

	interval = (double) delay_num / 10;
	end_num = 15 * interval;

	delay_num /= 2;

	delay_num = (int)(delay_num + interval);
	one_phase_probing();
	sleep(PhaseSleep);
	b_bw = get_bottleneck_bw((struct pkt_rcd_t *)&dst_gap, total_count);

	while (delay_num <= end_num) {
	    delay_num = (int)(delay_num + interval);
	    one_phase_probing();

	    if ((my_a_bw == 0) 
		&& (gap_comp(avr_dst_gap, avr_src_gap) == 0)) {
		my_a_bw = get_bandwidth();
	 	if (trace_fp != NULL) {
	    	    fprintf(trace_fp, "%%Seqential mode: get the value when delay_num = %d \n\n", delay_num); 
		}
		break;
	    }

	    sleep(PhaseSleep);
	}

	if (my_a_bw == 0) {
	    do {
	    	delay_num = (int)(delay_num + interval);
	    	one_phase_probing();
	        sleep(PhaseSleep);
	    } while (gap_comp(avr_dst_gap, avr_src_gap) != 0);
	    my_a_bw = get_bandwidth();
	    if (trace_fp != NULL) {
	        fprintf(trace_fp, 
		"%%Seqential mode: get the value when delay_num = %d \n\n", 
		delay_num); 
	    }
	}

	dump_bandwidth();

	probing_end_time = get_time();
	dump_stat();

	return my_a_bw;
}


int main(int argc, char *argv[]) 
{
	int opt;
	struct sockaddr_in server;
	struct sockaddr_in src_addr;
	int src_size;
	char cur_str[16];
	int result;

	char msg_buf[64];
	int msg_len;

	strcpy(work_mode, "sequential");

        while ((opt = getopt(argc, argv, "l:n:m:s:p:f:dhv")) != EOF) {
            switch ((char)opt) {
            case 'd':
		debug = 1;
                break;
            case 'm':
		strcpy(work_mode, optarg);
		if ((strcmp(work_mode, "binary") != 0) 
		    && (strcmp(work_mode, "fast") != 0) 
		    && (strcmp(work_mode, "sequential") != 0) 
		    && (strcmp(work_mode, "simple") != 0)) {
		    printf("Unknow working mode: %s. exit\n", work_mode);
		    exit(1);
		}
                break;
            case 'l':
		delay_num = atoi(optarg);
                break;
            case 'n':
		probe_num = atoi(optarg);
                break;
            case 's':
		packet_size = atoi(optarg);
                break;
            case 'p':
		dst_port = atoi(optarg);
                break;
            case 'v':
		verbose = 1;
                break;
            case 'f':
		if ((trace_fp = fopen(optarg, "w")) == NULL) {
		    printf("tracefile open fails, exits\n");
		    exit(1);
		}
                break;
	    case 'h':
	    default:
		Usage();
            }
	}
	switch (argc - optind) {
        case 1:
	    strcpy(dst, argv[optind]);
	    break;
	default:
	    Usage();
	}

	/* Setup signal handler for cleaning up */
	(void)setsignal(SIGTERM, cleanup);
	(void)setsignal(SIGINT, cleanup);
	if ((oldhandler = setsignal(SIGHUP, cleanup)) != SIG_DFL) 
	    (void)setsignal(SIGHUP, oldhandler);

  	/* Get hostname and IP address of target host */
  	if (GetCannonicalInfo(dst, dst_hostname, &dst_ip, dst_ip_str) < 0) {
    	    Quit();
  	}

        server.sin_addr.s_addr = dst_ip;
        server.sin_family = AF_INET;

	/* search for the right dst port to connect with */
        control_sock = socket(AF_INET, SOCK_STREAM, 0);
       	++dst_port;
        server.sin_port = htons(dst_port);
        result = connect(control_sock, (struct sockaddr *)&server, 
			sizeof(server));
   	while ((result < 0) && (dst_port < END_PORT)) {
	    close(control_sock);
            control_sock = socket(AF_INET, SOCK_STREAM, 0);
       	    ++dst_port;
            server.sin_port = htons(dst_port);
            result = connect(control_sock, (struct sockaddr *)&server, 
				    sizeof(server));
	    if (verbose) {
	        printf("dst_port = %d\n", dst_port);
	    }
   	};
   	if (result < 0) {
       	    perror("no dst port found, connection fails\n");
	    Quit();
	}

	/* Get hostname and IP address of source host */

	src_size = sizeof(struct sockaddr);
	getsockname(control_sock, (struct sockaddr *)&src_addr, &src_size);
	strcpy(src, inet_ntoa(src_addr.sin_addr));

	if (GetCannonicalInfo(src, src_hostname, &src_ip, src_ip_str) < 0) {
	    printf("get source address info fails, exit\n");
	    Quit();
	}

	if (verbose) {
	    printf("src addr: %s\n", src_ip_str);
	    printf("dst addr: %s\n", dst_ip_str);
	}

        /* send START message */
	sprintf(msg_buf, "$START$%d$", probe_num);
	msg_len = strlen(msg_buf);
        write(control_sock, msg_buf, msg_len);

        /* wait for READY ack */
	get_item(control_sock, cur_str);
	if (verbose) {
	    printf("we get str: %s\n", cur_str);
	}
        if (strncmp(cur_str, "READY", 5) != 0) {
	    printf("get unknow str when waiting for READY from dst, exit\n");
	    Quit();
        }

	get_item(control_sock, cur_str);
	if (verbose) {
	    printf("probing_port = %s\n", cur_str);
	}
	probing_port = atoi(cur_str);

	if (init_session(src_ip, src_port, dst_ip, dst_port) == 0) {
	    printf("ERROR: Couldn't establish session to %s:%d\n",
   			dst_hostname, dst_port);
	    Quit();
	}

	/* allow dst 2 seconds to start up the packet filter */
	sleep(2);
        /* probing */
	if (delay_num > 0) {
	    one_phase_probing();
	    get_bandwidth();
	    dump_bandwidth();
	}
	else if (strcmp(work_mode, "binary") == 0) 
            binary_probing();
	else if (strcmp(work_mode, "fast") == 0) 
            fast_probing();
	else if (strcmp(work_mode, "sequential") == 0) 
            sequential_probing();
	else if (strcmp(work_mode, "simple") == 0) 
            simple_probing();


	/* finishing */
  	cleanup(1);

	return (0);
}
